## Tree Diff
## =========
##
## Utility to compare the file contents of two directory trees.

import std/[json, jsonutils, os, tables, sequtils, strutils]
import docopt
import incremental_md5, console_progress

import ./cliconstants


type
  ProgressWrapper* = tuple[impl: Progress, verbosity: Verbosity]
    ## Wrapper around a console_progress.Progress.

  Verbosity* = enum ## Enum representing the level of output verbosity the tool will emit.
    very_quiet, ## suppress all output including the progress indicator
    quiet,      ## suppress all output except the progress indicator
    normal      ## emit all output

proc newProgressWrapper*(outFile = stdout, verbosity = normal): ProgressWrapper =
  ## Create a new ProgressWrapper for the given verbosity.
  if verbosity > very_quiet:
    result = (impl: newProgress(0, outFile), verbosity: verbosity)
  else: result = (impl: nil, verbosity: verbosity)

proc init(p: ProgressWrapper, root: string, fileCount: int): void =
  if p.verbosity == normal:
      echo "-- ", root.expandFilename, "\L   ", fileCount, " files"
  if p.verbosity > very_quiet: p.impl.setMax(fileCount)

proc update(p: ProgressWrapper, count: int, file: string): void =
  if p.verbosity > very_quiet:
    p.impl.updateProgress(count, file[max(file.high - 15, 0)..file.high])

proc finish(p: ProgressWrapper): void =
  if p.verbosity > very_quiet:
    p.impl.erase
  if p.verbosity == normal: echo "  ", p.impl.getMax, " files.\L"



proc countFiles(root: string): int =
  for file in walkDirRec(root): result += 1

proc getRelPath(ancestor, child: string): string =
  ## Given a ancestor path and a child path, assuming the child path is
  ## contained within the ancestor path, return the relative path from the
  ## ancestor to the child.

  let ancestorPath = ancestor.expandFilename.split({DirSep, AltSep})
  let childPath = child.expandFilename.split({DirSep, AltSep})

  # If the ancestor path is longer it cannot contain the child path and we
  # cannot construct a relative path without backtracking.
  if (ancestorPath.len > childPath.len): return ""

  # Compare the ancestor and child path up until the end of the ancestor path.
  var idx = 0
  while idx < ancestorPath.len and ancestorPath[idx] == childPath[idx]: idx += 1

  # If we stopped before reaching the end of the ancestor path it must be that
  # the paths do not match. The ancestor cannot contain the child and we cannot
  # build a relative path without backtracking.
  if idx != ancestorPath.len: return ""
  return foldl(@["."] & childPath[idx..childPath.high], joinPath(a, b))




type
  FileEntry* = ref tuple[relPath: string, checksum: string]
    ## Data about one file that has been analyzed

  DirAnalysis* = ## Analysis data about one directory tree.
    tuple[allEntries: seq[FileEntry],
          byRelPath: TableRef[string, FileEntry],
          byChecksum: TableRef[string, seq[FileEntry]]]


  DisplayOptions = tuple[left, right, same, content, path: bool]
    ## Consolidated description of which types of results to display.

func `$`(f: FileEntry): string = f.checksum & ": " & f.relPath

proc getOrFail(n: JsonNode, key: string, objName: string = ""): JsonNode =
  ## convenience method to get a key from a JObject or raise an exception
  if not n.hasKey(key): raise newException(Exception, objName & " missing key '" & key & "'")
  return n[key]

proc getIfExists(n: JsonNode, key: string): JsonNode =
  ## convenience method to get a key from a JObject or return null
  result = if n.hasKey(key): n[key]
           else: newJNull()

func parseFileEntry(n: JsonNode): FileEntry =
  result = new(FileEntry)
  result.relPath = n.getOrFail("relPath").getStr
  result.checksum = n.getOrFail("checksum").getStr

func initDirAnalysis(): DirAnalysis =
  (allEntries: @[],
   byRelPath: newTable[string, FileEntry](),
   byChecksum: newTable[string, seq[FileEntry]]())

func indexEntries(da: var DirAnalysis) =
  for e in da.allEntries:
    da.byRelPath[e.relPath] = e
    if not da.byChecksum.hasKey(e.checksum):
      da.byChecksum[e.checksum] = newSeq[FileEntry]()
    da.byChecksum[e.checksum].add(e)

proc analyzeDir*(root: string, progress: ProgressWrapper): DirAnalysis =
  ## Inspect a directory and analyze all files, noting their relative paths and
  ## checksum of their contents.
  let fileCount = countFiles(root)

  progress.init(root, fileCount + 10)

  result = initDirAnalysis()

  var count = 0
  for file in walkDirRec(root):
    let md5sum = fileToMd5(file)

    var fileEntry: FileEntry = new(FileEntry)
    fileEntry[] = (relPath: getRelPath(root, file), checksum: md5sum)
    result.allEntries.add(fileEntry)

    progress.update(count, file)
    count += 1

  result.indexEntries
  count += 10
  progress.finish()

proc loadAnalysis*(path: string): DirAnalysis =
  ## Load a previously performed directory analysis.
  let allEntriesJson = parseJson(readFile(path))
  result = initDirAnalysis()
  result.allEntries = toSeq(items(allEntriesJson)).map(parseFileEntry)
  result.indexEntries

proc saveAnalysis*(path: string, analysis: DirAnalysis): void =
  ## Save a completed analysis.
  writeFile(path, $(analysis.allEntries.toJson))

proc intersection*(left, right: DirAnalysis): seq[FileEntry] =
  ## Find all ``FileEntry`` that are the same on both sides: matching contents
  ## and paths.
  return left.allEntries.filter do (item: FileEntry) -> bool:
    if not right.byRelPath.hasKey(item.relPath): return false
    let match = right.byRelPath[item.relPath]
    if match == nil: return false
    return item.checksum == match.checksum

proc difference*(left, right: DirAnalysis): seq[FileEntry] =
  ## Find all ``FileEntry`` that are present in the left but not present in
  ## the right.
  return left.allEntries.filter do (item: FileEntry) -> bool:
    return not right.byRelPath.hasKey(item.relPath) and
           not right.byChecksum.hasKey(item.checksum)

proc `*`*(left, right: DirAnalysis): seq[FileEntry] {.inline.} =
  ## Alias for `intersection(left, right) <#intersection>`_
  return intersection(left, right)

proc `-`*(left, right: DirAnalysis): seq[FileEntry] {.inline.} =
  ## Alias for `difference(left, right) <#difference>`_
  return difference(left, right)

proc samePathDifferentContents*(left, right: DirAnalysis): seq[string] =
  ## Find all ``FileEntry`` that have the same paths in both trees but whose
  ## contents differ.
  let matchingEntries = left.allEntries.filter do (item: FileEntry) -> bool:
    if not right.byRelPath.hasKey(item.relPath): return false
    let match = right.byRelPath[item.relPath]
    return item.checksum != match.checksum
  return matchingEntries.map(proc(item: FileEntry): string = return item.relPath)

proc sameContentsDifferentPaths*(left, right: DirAnalysis): seq[tuple[left, right: FileEntry]] =
  ## Find all ``FileEntry`` whose contents are the same in both trees but
  ## which are located at differenc paths.
  result = @[]
  for item in left.allEntries:
    if not right.byChecksum.hasKey(item.checksum): continue
    for match in right.byChecksum[item.checksum]:
      if item.relPath != match.relPath: result.add((left: item, right:match))

when isMainModule:

  let quitWithError = proc (error: string): void =
    stderr.writeLine("treediff: " & error)
    quit(QuitFailure)

  let args = docopt(USAGE, version = "treediff " & VERSION)

  var verbosity = normal
  if args["--quiet"]: verbosity = quiet
  if args["--very-quiet"]: verbosity = very_quiet
  let progressWrapper = newProgressWrapper(verbosity = verbosity)

  # Load or perform analysis
  if not args["<left>"]:
    quitWithError("Missing <left> parameter.")

  let leftPath: string = $args["<left>"]

  let loadPath = proc (path: string): DirAnalysis =
    if not path.fileExists and not path.dirExists:
      quitWithError($path & ": no such file or directory.")

    let fileInfo = path.getFileInfo
    if fileInfo.kind == pcDir:
      return analyzeDir(path, progressWrapper)
    elif fileInfo.kind == pcFile:
      result = loadAnalysis(path)
    else:
      quitWithError($path & ": is not a file or directory")

  var leftAnalysis, rightAnalysis: DirAnalysis

  leftAnalysis = loadPath(leftPath)

  if not args["<right>"]:
    rightAnalysis = (allEntries: @[],
                 byRelPath: newTable[string, FileEntry](),
                 byChecksum: newTable[string, seq[FileEntry]]())
  else:
    var rightPath: string = $args["<right>"]
    rightAnalysis = loadPath(rightPath)

  # Check for output options
  if args["--save-left"]:
    saveAnalysis($args["--save-left"], leftAnalysis)

  if args["--save-right"] and rightAnalysis.allEntries.len > 0:
    saveAnalysis($args["--save-right"], rightAnalysis)

  # Parse filter options
  var displayOptions: DisplayOptions = (
    left: false, right: false, same: false, content: false, path: false)

  # If none of the explicit selectors are given, assume all are expected.
  if not (args["--left-only"] or args["--right-only"] or
          args["--same"] or args["--content-mismatch"] or
          args["--path-mismatch"] ):
    displayOptions = (left: true, right: true, same: true,
                      content: true, path: true)

  if args["--same"]:                      displayOptions.same = true
  if args["--exclude-same"]:              displayOptions.same = false
  if args["--content-mismatch"]:          displayOptions.content = true
  if args["--exclude-content-mismatch"]:  displayOptions.content = false
  if args["--path-mismatch"]:             displayOptions.path = true
  if args["--exclude-path-mismatch"]:     displayOptions.path = false
  if args["--left-only"]:                 displayOptions.left = true
  if args["--exclude-left-only"]:         displayOptions.left = false
  if args["--right-only"]:                displayOptions.right = true
  if args["--exclude-right-only"]:        displayOptions.right = false

  # Display output results
  if verbosity == normal:
    if displayOptions.same:
      let sameEntries = leftAnalysis * rightAnalysis
      for fe in sameEntries: echo "same: ", fe.relPath

    if displayOptions.content:
      let contentsDiffer = samePathDifferentContents(leftAnalysis, rightAnalysis)
      for path in contentsDiffer: echo "contents differ: ", path

    if displayOptions.path:
      let pathsDiffer = sameContentsDifferentPaths(leftAnalysis, rightAnalysis)
      for pair in pathsDiffer:
        echo "paths differ: ", pair.left.relPath, " ", pair.right.relPath

    if displayOptions.left:
      let leftOnly = leftAnalysis - rightAnalysis
      for fe in leftOnly: echo "left only: ", fe.relPath

    if displayOptions.right:
      let rightOnly = rightAnalysis - leftAnalysis
      for fe in rightOnly: echo "right only: ", fe.relPath
